use crate::cpu::Word;

use super::Section;

#[derive(Debug)]
pub enum ParseError {
    EmptyHeader,
    MagicNumberCheckFail,
    UnexpectedHeaderEnd,
    UnexpectedFileEnd,
    Utf8ConvError,
}

/// Reads a requence of u16, and returns a rust utf8 owned string
/// and the index of the next 16bit word that follows the string,
/// so the first 16 bit word after the null delimiter.
/// It splits every u16 into two chunks of 8 bits, reads until
/// it finds the first empty u8 and attempts to convert the u8 array to
/// a rust UTF8 String.
pub fn find_and_read_string(s: &[u16]) -> Result<(String, usize), ParseError> {
    let mut bytes = vec![];

    let mut index: usize = 0;
    for b in s.iter() {
        let x0 = (*b & 0xFF00) >> 8;
        let x1 = *b & 0x00FF;
        // exit when the first 0 bit is found.
        if x0 == 0 {
            index += 1;
            break;
        };
        bytes.push(x0 as u8);
        if x1 == 0 {
            index += 1;
            break;
        };
        bytes.push(x1 as u8);
        index += 1;
    }

    let s = String::from_utf8(bytes).map_err(|_| ParseError::Utf8ConvError)?;

    return Ok((s, index));
}

/// Takes a binary file and returns a list of sections
pub fn read_binary(b: &[u16]) -> Result<Vec<Section>, ParseError> {
    let mut res = vec![];

    let headers = parse_header(b)?;
    let hlen = headers.len() * 2 + 2; // two 16bits words for every entry,
                                      // and 2 etxra 16 bits number at the start

    for (offset, length) in headers {
        // section start. The name begins here
        let start = hlen + offset as usize;
        let str_buffer = b.get(start..).ok_or(ParseError::UnexpectedHeaderEnd)?;
        let (name, i) = find_and_read_string(str_buffer)?;

        let c_start = start + i;
        let c_end = start + (length as usize) + i;

        println!("{:?}, start: {}, end: {}", b, c_start, c_end);
        let Some(content) = b.get((c_start)..(c_end)) else {return Err(ParseError::UnexpectedFileEnd)};

        res.push(Section::new(name, content))
    }

    Ok(res)
}

/// Parses binary headers
fn parse_header(b: &[Word]) -> Result<Vec<(u16, u16)>, ParseError> {
    let Some([_m, s]) = b.get(0..2) else {return Err(ParseError::EmptyHeader)};

    // Magic number check. Can go unchecked, check spec.
    // if (*m != MAGIC) {
    //     return Err(WrongMagicNum)
    // };

    // s is the number of pairs (offset, length) in our header.
    // since we're counting pairs, we need s*2 numbers from input.

    let Some(headerdata) = b.get(2..((*s as usize * 2) + 2)) else {return Err(ParseError::UnexpectedHeaderEnd)};

    assert!(
        headerdata.len() % 2 == 0,
        "Header does not have an even number of words"
    );

    let mut hd = headerdata.iter();

    let mut res = vec![];

    while let (Some(offset), Some(len)) = (hd.next(), hd.next()) {
        res.push((*offset, *len))
    }

    Ok(res)
}
