use super::constants::MAGIC;

use super::Section;

impl Section<'_> {
    /// Converts the entry's name to utf8 packed in bits of length 16.
    fn serialize_name(&self) -> Vec<u16> {
        return make_string(&self.name);
    }

    /// Joins the name of the section followed by its contents in a
    /// vector of 16 bits.
    fn serialize(&self) -> Vec<u16> {
        let mut tmp = make_string(&self.name);
        tmp.append(&mut self.content.to_owned());
        return tmp;
    }
}

#[derive(Debug)]
/// Entry in the symbols table.
/// Consists of an offset and a lenght.
struct STEntry {
    offset: u16,
    length: u16,
}

/// Takes a string, and creates a u16, null-terminated utf8 string,
/// in a vector of u16 (padding possible)
pub fn make_string(s: &str) -> Vec<u16> {
    let raw_bytes: &[u8] = s.as_bytes();

    let mut rb = raw_bytes.iter();
    // raw_bytes must be converted to u16.
    //
    let mut bytes: Vec<u16> = {
        let mut res = vec![];
        // highly cursed: depends on the order in which the arguments of a tuple are
        // evaluated. Does its job!
        while let (Some(word0), Some(word1)) = (rb.next(), rb.next()) {
            // println!("Pair: {}, {}, word: {:?}", word0, word1, raw_bytes);
            res.push(((*word0 as u16) << 8) + (*word1 as u16));
        }
        // if we branch into this else, either there's a single word left or zero.
        // since, in case a single word was left, the first rb.next() call in the line
        // above would've consumed it, we have to gather that last element again
        match raw_bytes.len() {
            0 => res,
            n => {
                if n % 2 != 0 {
                    // if there's an uneven number of chunks of 8 bits,
                    // we introduce padding!
                    // println!("Adding last one too");
                    res.push((*raw_bytes.last().unwrap() as u16) << 8);
                }
                res
            }
        }
    };

    // adding null termination byte
    bytes.push(0);

    return bytes;
}

/// Takes a list of Section struct to be inserted in the symbols table and returns
/// both the table and the inserted data, with offsets
fn conv(sy_table: &[Section]) -> (Vec<STEntry>, Vec<u16>) {
    let mut current_offset: u16 = 0;
    let mut entries: Vec<STEntry> = vec![];
    let mut content: Vec<u16> = vec![];

    for entry in sy_table {
        let mut binary_entry = entry.serialize();

        // Add name + content to the whole file
        content.extend_from_slice(&mut binary_entry);

        // take note of the current offset and content lenght in the entry table
        // (without including the length of the string)
        entries.push(STEntry {
            offset: current_offset,
            length: entry.content.len() as u16,
        });

        // add to the current offset the length of the data we've saved:
        current_offset += binary_entry.len() as u16;
    }

    return (entries, content);
}

/// Takes a list of STentry and serializes them
fn make_header(sy_table: &[STEntry]) -> Vec<u16> {
    let mut res = vec![];
    for entry in sy_table {
        res.push(entry.offset);
        res.push(entry.length);
    }

    res
}

/// Takes a list of Sections and returns a binary file.
pub fn make_binary(sections: &[Section]) -> Vec<u16> {
    let (sy_table, mut data) = conv(sections);

    let mut header = make_header(&sy_table);

    let mut res = vec![MAGIC, sections.len() as u16];
    res.append(&mut header);
    res.append(&mut data);
    res
}
